#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from poppy.core.logger import logger

from functools import wraps
from poppy.core.generic.metaclasses import SingletonManager
from poppy.core.db.database import Database
from poppy.core.db.dry_runner import DryRunner



__all__ = ["Connector"]


class ConnectorNotAvailable(Exception):
    pass


class Connector(object, metaclass=SingletonManager):
    """
    Base class for managing sessions with various databases.
    """
    def __init__(self, name, database=None):
        """
        Init a connector with useful information.
        """
        # store the name of the connector, useful to reference it later
        self.name = name

        # init the session to a None value to see if already in use or not
        self.session = None

        # if a database is given, set it
        if database is not None:
            self.database = database

    @DryRunner.dry_run
    def bind(self):
        """
        To bind the database to the models of the ORM.
        """
        # get the database
        database = self.get_database()

        # and connect
        database.connectDatabase()

        # check the connection is good
        if not database.connected:
            message = "{0} is not connected".format(database)
            logger.error(message)
            raise Exception(message)

        # check that a session is already in use
        if self.session is not None:
            # close the session
            try:
                self.session.close()
            except Exception as e:
                logger.error(e)
                raise e

        # get a scoped session to be synchronized with others
        self.session = self.factory(database)

    def rollback(self):
        """
        Undo what has been done in the current transaction.
        """
        self.session.rollback()

    @DryRunner.dry_run
    def update_database(self):
        """
        Commit all changes for the current session.
        """
        self.session.commit()

    @DryRunner.dry_run
    def flush(self):
        """
        Flush the session, i.e. it transfers python objects to the transaction
        buffer of the database, thus the memory of the session is cleaned up.
        But if an error occurred, the transaction is rolled back.
        """
        self.session.flush()

    @DryRunner.dry_run
    def add(self, obj):
        """
        Add an object into the session. A simple wrapper around the session.
        """
        # add the object into the shared session
        self.session.add(obj)

    def factory(self, database):
        """
        Given the database, return the session factory.
        """
        return database.session_factory()

    @property
    def database(self):
        return self._database

    @database.setter
    def database(self, database):
        # store database in use
        self._database = database

    def get_database(self):
        """
        Return the database object.
        """
        return Database.manager[self._database]

    @classmethod
    def if_connected(cls, *args):
        """
        A decorator to execute a function or method only if the database named
        name is connected.
        """
        def decorator(func):
            """
            Return the function that will be used to decorate.
            """
            @wraps(func)
            def wrapper(*fargs, **kwargs):
                """
                The function that will check the connection and run the wrapped
                function if the database is available.
                """
                # loop over names
                for name in args:
                    # get the connector instance from the given name
                    if name not in cls.manager:
                        message = "Connector {0} not available".format(name)
                        logger.error(message)
                        raise ConnectorNotAvailable(message)

                    # get the connector
                    connector = cls.manager[name]

                    # get the database
                    database = connector.get_database()

                    # check the connection
                    if not database.is_available_with_error():
                        message = "No connection for {0}".format(database)
                        return None

                # run the wrapped function as usual
                return func(*fargs, **kwargs)

            return wrapper

        return decorator

    def __repr__(self):
        """
        A better representation of a connector.
        """
        return "Connector {0}".format(self.name)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import re


def get_valid_filename(s):
    """
    Returns the given string converted to a string that can be used for a clean
    filename. Specifically, leading and trailing spaces are removed; other
    spaces are converted to underscores; and anything that is not a unicode
    alphanumeric, dash, underscore, or dot, is removed.
    >>> get_valid_filename("john's portrait in 2004.jpg")
    'johns_portrait_in_2004.jpg'
    """
    text = s.strip().replace(' ', '_')
    return re.sub(r'(?u)[^-\w.]', '', text)


# vim: set tw=79 :



import logging
import importlib
import os
import shutil
import stat
import sys
from datetime import datetime
from os import path
from pprint import pformat
import declic as cli
from sqlalchemy import UniqueConstraint

import jinja2
import poppy.core as poppy_core

logger = logging.getLogger('poppy-cli')


def render(template_filepath, context):
    template_path, filename = os.path.split(template_filepath)
    return jinja2.Environment(
        loader=jinja2.FileSystemLoader(template_path or './')
    ).get_template(filename).render(context)


def make_writeable(filename):
    """
    Make sure that the file is writeable.
    Useful if our source is read-only.
    """
    if sys.platform.startswith('java'):
        # On Jython there is no os.access()
        return
    if not os.access(filename, os.W_OK):
        st = os.stat(filename)
        new_permissions = stat.S_IMODE(st.st_mode) | stat.S_IWUSR
        os.chmod(filename, new_permissions)


def create_from_template(pipeline_or_plugin, name, directory, namespace=None):
    """
    Create pipeline/plugin directory tree from POPPy templates

    :param pipeline_or_plugin: String indicating which template to use ('pipeline' or 'plugin')
    :param name: name of the pipeline/plugin
    :param directory: path to the plugin/pipeline directory
    :return:
    """
    base_subdir = '%s_template' % pipeline_or_plugin

    # store the template dir path
    template_dir = path.join(poppy_core.__path__[0], 'management', 'templates', base_subdir)

    prefix_length = len(template_dir) + 1

    # if some directory is given, make sure it's nicely expanded
    if directory is None:
        top_dir = path.join(os.getcwd(), name)
        try:
            os.makedirs(top_dir)
        except FileExistsError:
            raise FileExistsError("'%s' already exists" % top_dir)
        except OSError as e:
            raise OSError(e)
    else:
        top_dir = os.path.abspath(path.expanduser(directory))
        if not os.path.exists(top_dir):
            raise FileNotFoundError("Destination directory '%s' does not "
                                    "exist, please create it first." % top_dir)

    base_name = '%s_name' % pipeline_or_plugin
    base_directory = '%s_directory' % pipeline_or_plugin
    camel_case_name = 'camel_case_%s_name' % pipeline_or_plugin
    camel_case_value = ''.join(x for x in name.title() if x != '_')
    databases = [
        {
            "identifier": "MAIN-DB",
            "short_identifier": 'MAIN-DB',
            "vendor": "postgresql",
            "connector": "poppy.pop.db_connector.POPPy",
            "name": "poppy_tuto",
            "host": "localhost",
            "admin_account": "pipeadmin",
            "admin_password": "adminpwd",
            "user_account": "pipeuser",
            "user_password": "userpwd",
            "short_description": "POPPy tutorial database",
            "description": "POPPy tutorial Database",
            "release": {
                       "author": "POPPy team",
                       "date": "2018-06-26",
                       "institute": "LESIA",
                       "reference": "POPPy_Framework_User_Manual",
                       "version": "0.1.0",
                       "modification": ""
                       }
        }
    ]

    context = {
        base_name: name,
        base_directory: top_dir,
        camel_case_name: camel_case_value,
        'plugin_namespace': namespace,
        'poppy_core_version': poppy_core.__version__,
        'provider': 'LESIA',
        'release_date': datetime.utcnow().strftime('%Y-%m-%d'),
        'databases': databases,
        'databases_with_description': [x for x in databases if 'description' in x]
    }

    logger.debug('context: %s' % pformat(context))

    for root, dirs, files in os.walk(template_dir):

        path_rest = root[prefix_length:]

        # create relative filepath
        if namespace is not None:
            relative_dir = path_rest.replace('plugin_namespace', namespace).replace(base_name, name)
        else:
            relative_dir = path_rest.replace(base_name, name)
        if relative_dir:
            target_dir = path.join(top_dir, relative_dir)
            if not path.exists(target_dir):
                os.mkdir(target_dir)

        # ignore pycache and hidden directories
        for dir_name in dirs[:]:
            if dir_name.startswith('.') or dir_name == '__pycache__':
                dirs.remove(dir_name)

        for filename in files:
            if filename.endswith(('.pyo', '.pyc', '.py.class')):
                # ignore some files as they cause various breakages.
                continue
            old_path = path.join(root, filename)
            if filename[-4:] == '-tpl':
                new_filename = filename[:-4]
            else:
                new_filename = filename

            new_path = path.join(top_dir, relative_dir,
                                 new_filename.replace(base_name, name))

            if path.exists(new_path):
                raise FileExistsError("%s already exists, overlaying a "
                                      "project or app into an existing "
                                      "directory won't replace conflicting "
                                      "files" % new_path)

            if filename[-4:] == '-tpl':
                content = render(old_path, context)
                with open(new_path, 'w', encoding='utf-8') as new_file:
                    new_file.write(content)
            else:
                shutil.copyfile(old_path, new_path)

                logger.debug("Creating %s\n" % new_path)
            try:
                shutil.copymode(old_path, new_path)
                make_writeable(new_path)
            except OSError:
                logger.error("Notice: Couldn't set permission bits on %s. You're "
                             "probably using an uncommon filesystem setup. No "
                             "problem." % new_path)


def setup_logger(quiet, debug):

    # set logger formatter
    formatter = logging.Formatter('%(asctime)s | %(name)s | %(levelname)s | %(message)s')
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    # set logging level
    if debug:
        logger.setLevel(logging.DEBUG)
    elif quiet:
        logger.setLevel(logging.WARNING)
    else:
        logger.setLevel(logging.INFO)

    logging_level = logging.getLevelName(logger.getEffectiveLevel())
    logger.debug('Logging level: %s' % logging_level)


@cli.group(on_before=setup_logger)
@cli.argument('--version', action='version', version=poppy_core.__version__)
@cli.argument('-q', '--quiet', action='store_true')
@cli.argument('-d', '--debug', action='store_true')
def poppy_cli():
    pass


@poppy_cli.group()
def create():
    pass


@create.command(description='Create a new pipeline', chain=True)
@cli.argument('name', help='Name of the pipeline')
@cli.argument('--directory', help='Optional destination directory')
def pipeline(name, directory):
    create_from_template('pipeline', name, directory)


@create.command(description='Create a new plugin', chain=True)
@cli.argument('name', help='Name of the plugin')
@cli.argument('--directory', help='Optional destination directory')
@cli.argument('--namespace', help='Plugin namespace')
@cli.argument('--cython', action='store_true', help='Create a Cython-ready plugin')
def plugin(name, directory, namespace, cython):
    if name.count('.') > 1:
        raise ValueError(f'{name} has more than one \'dot\' character')

    if '.' in name:
        split_name = name.split('.')
        namespace = split_name[0]
        name = split_name[1]

    if namespace is None:
        # FIXME
        raise NotImplementedError('Plugins without namespace are not yet supported')

    create_from_template('plugin', name, directory, namespace)


@poppy_cli.command(description='Generate the database documentation of a plugin')
@cli.argument('plugin_name', help='The name of the plugin we want to generate the documentation of')
def gen_doc(plugin_name):
    """
    This function generate an reST file in the current directory. This file
    contains the model documentation about a given plugin.
    """

    plugin_models = importlib.import_module(f'{plugin_name}.models')
    try:
        tables = plugin_models.tables
    except AttributeError:
        raise AttributeError('Are you sure you have included the mandatory code '
                             'to your plugin.models.__init__.py ? See the '
                             'poppy documentation about models.')

    file_name = f'{plugin_name}_model_documentation.rst'
    file_path = os.path.join(os.path.curdir, file_name)

    csv_table_header = '.. csv-table:: {0}\n' \
                       '   :header: "Column name", "Data type", "Description", ' \
                       '"Priority", "Comment"\n\n'

    with open(file_path, 'w') as f:

        # A dictionary containing schema name as key, and all the generated text
        # of the tables in each schema as value
        table_doc_dict = dict()

        for table_name, model_class in tables.items():

            text = ''
            schema_name = 'public'

            # Title of the section
            title = f'The table {table_name}\n'
            text += title
            text += f"{'='*(len(title)-1)}\n"

            # The docstring of the class representation of the table
            text += ' '.join(model_class.__doc__.replace('\n', '').split()) # Replaces sequences of space with one and remove linebreaks
            text += '\n\n'

            text += csv_table_header.format(table_name)

            for c in model_class.__table__.columns:
                col_infos = c.infos()
                text += f'   "{col_infos["name"]}", "{col_infos["sql_type"]}", "{col_infos["description"]}", "{col_infos["priority"]}", "{col_infos["comment"]}"\n'

            text += '\n'

            table_args = model_class.__table_args__

            # Extract unique constraints and schema name (if any)
            if type(table_args) == tuple:
                for arg in table_args:
                    if type(arg) == UniqueConstraint:
                        constraint = arg
                        columns = '(' + ','.join([x.key for x in constraint]) + ')'
                        text += f'The tuple of columns {columns} must be unique.\n'
                    elif type(arg) == dict:
                        try:
                            schema_name = arg['schema']
                        except KeyError:
                            pass

            elif type(table_args) == dict:
                try:
                    schema_name = table_args['schema']
                except KeyError:
                    pass

            text += '\n'

            try:
                table_doc_dict.update({schema_name:table_doc_dict[schema_name]+text})
            except KeyError:
                table_doc_dict.update({schema_name:text})


        for schema_name, text in table_doc_dict.items():

            schema_header = f'The \'{schema_name}\' schema\n'

            f.write(f"{'*'*(len(schema_header)-1)}\n")
            f.write(schema_header)
            f.write(f"{'*'*(len(schema_header)-1)}\n\n")

            f.write(table_doc_dict[schema_name])



    logger.info(f'{file_name} successfully generated in the current directory')


if __name__ == '__main__':
    poppy_cli()



# -*- coding: utf-8 -*-
import collections
import json
import os

ROOT_DIRECTORY = os.environ.get('PIPELINE_ROOT_DIRECTORY', '')

MAIN_DATABASE = 'MAIN-DB'

PLUGINS = [
    'poppy-pop',
]

# load the logger config
conf_dir_path = os.path.dirname(os.path.realpath(__file__))

logger_json_filepath = os.path.join(conf_dir_path, 'logger.json')

with open(logger_json_filepath, 'rt') as json_file:
    LOGGER_CONFIG = json.load(json_file, object_pairs_hook=collections.OrderedDict)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from poppy.core.logger import logger

from .base_task import BaseTask


class Task(BaseTask):
    _targets_definition = None

    @classmethod
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        # reset the targets definition attribute to avoid sharing targets definition with subclasses
        cls._targets_definition = {'inputs': [], 'outputs': []}

    @classmethod
    def legacy_as_task(cls, run_func=None, *, plugin_name=None, name=None, legacy=True):
        """
        Method used to ensure backward compatibility with the previous decorator API

        :param run_func: the decorated function
        :param plugin_name: the plugin name
        :param name: the task name
        :param legacy: flag indicating which API is used (the legacy or the new one)
        :return: a task class that run the decorated function
        """
        _plugin_name = cls.plugin_name if plugin_name is None else plugin_name
        _name = cls.name if name is None else name

        if legacy:
            logger.warning(f"Deprecation Warning: Task ('{_plugin_name}/{_name}') "
                           "should be declare with the new decorator API")

        return type(
            cls.to_camel_case(_name),
            (cls,),
            {'plugin_name': _plugin_name, 'name': _name, 'run': run_func}
        )

    @staticmethod
    def to_camel_case(snake_str):
        components = snake_str.split('_')
        # We capitalize the first letter of each component and join them together
        return ''.join(x.title() for x in components)

    @classmethod
    def as_task(cls, run_func=None, *, plugin_name=None, name=None):
        """
        Use the decorated function as a task

        :param run_func: the decorated function
        :param plugin_name: the plugin name
        :param name: the task name
        :return: a task class that run the decorated function
        """

        if run_func is None:
            def inner_function(_run_func):
                return cls.legacy_as_task(_run_func, plugin_name=plugin_name, name=name, legacy=False)

            return_value = inner_function
        else:
            return_value = cls.legacy_as_task(run_func)

        return return_value

    def instantiate_target(self, target_list, target_definition):
        """
        Instantiate a new target using the given definition and store the target id in a list

        :param target_list: the list used to store the target ids
        :param target_definition: the target definition
        :return:
        """
        target_class, identifier, args, kwargs = target_definition

        target_class(identifier, self.pipeline, *args, **kwargs)

        target_list.append(identifier)


    def instantiate_targets(self):

        # get the targets defined in the task class
        self.add_targets()

        # instantiate each target
        for target_definition in self._targets_definition['inputs']:
            self.instantiate_target(self._inputs, target_definition)

        for target_definition in self._targets_definition['outputs']:
            self.instantiate_target(self._outputs, target_definition)

    def add_targets(self):
        """
        Method to override to add targets

        :return:
        """
        pass

    def __init__(self, instance_name=None):
        plugin_instance = self.get_plugin()
        task_descriptor = self.get_descriptor()

        super().__init__(plugin=plugin_instance, descriptor=task_descriptor, instance_name=instance_name)

        # store the id of the inputs/outputs targets
        self._inputs = []
        self._outputs = []

    @property
    def inputs(self):
        """
        Generate a subset of the pipeline targets containing only the task inputs

        :return: the input targets
        """
        return {target_id: self.pipeline.targets[target_id] for target_id in self._inputs}

    @property
    def outputs(self):
        """
        Generate a subset of the pipeline targets containing only the task outputs

        :return: the output targets
        """
        return {target_id: self.pipeline.targets[target_id] for target_id in self._outputs}

    @classmethod
    def get_descriptor(cls):
        plugin_instance = cls.get_plugin()
        return plugin_instance.get_task_descriptor(cls.name)

    @classmethod
    def get_plugin(cls):
        from poppy.core.plugin import Plugin
        return Plugin.manager[cls.plugin_name]


    @classmethod
    def add_input(cls, target_class, identifier, *args, **kwargs):
        # create the targets def dict if not defined
        cls._targets_definition['inputs'].append((target_class, identifier, args, kwargs))

    @classmethod
    def add_output(cls, target_class, identifier, *args, **kwargs):
        cls._targets_definition['outputs'].append((target_class, identifier, args, kwargs))

    def input(self):
        task_descriptor = self.get_descriptor()
        return task_descriptor["inputs"]

    def output(self):
        task_descriptor = self.get_descriptor()
        return task_descriptor["outputs"]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
.. module:: spacepy.pycdf.const

Various constants defined in cdf.h and used in pycdf.
Most constants referred to in the CDF manuals are provided by this module.
E.g., to create a CDF and add a variable of type EPOCH:

>>> from spacepy import pycdf
>>> cdf = pycdf.CDF('new.cdf', '')
>>> cdf.new('epoch', type=pycdf.const.CDF_EPOCH)

Copyright 2010-2012 Los Alamos National Security, LLC.
"""

import ctypes

# Generated by h2py from cdf.h
# Regeneration process:
# Edit a copy of cdf.h
# M-x replace-string (CDFstatus) with nothing
# h2py cdf.h
# Open CDF.py
# remove def lines
# M-x replace-string ( with nothing, also ) with nothing
# add RESERVED_CDFID = 0 before RESERVED_CDFSTATUS
# M-x replace-regexp \(-?[0-9]+\)L with: ctypes.c_long(\1)
# mark EPOCH as double: ILLEGAL_EPOCH_VALUE = ctypes.c_double(-1.0)

TT2000NULL = ctypes.c_double(0)
CDF_MIN_DIMS = 0
CDF_MAX_DIMS = 10
CDF_VAR_NAME_LEN = 64
CDF_ATTR_NAME_LEN = 64
CDF_VAR_NAME_LEN256 = 256
CDF_ATTR_NAME_LEN256 = 256
CDF_COPYRIGHT_LEN = 256
CDF_STATUSTEXT_LEN = 120
CDF_PATHNAME_LEN = 512
EPOCH_STRING_LEN = 24
EPOCH1_STRING_LEN = 16
EPOCH2_STRING_LEN = 14
EPOCH3_STRING_LEN = 24
EPOCH4_STRING_LEN = 23
EPOCH16_STRING_LEN = 36
EPOCH16_1_STRING_LEN = 24
EPOCH16_2_STRING_LEN = 14
EPOCH16_3_STRING_LEN = 36
EPOCH16_4_STRING_LEN = 32
TT2000_0_STRING_LEN = 30
TT2000_1_STRING_LEN = 19
TT2000_2_STRING_LEN = 14
TT2000_3_STRING_LEN = 29
EPOCHx_STRING_MAX = 50
EPOCHx_FORMAT_MAX = 68
CDF_INT1 = ctypes.c_long(1)
CDF_INT2 = ctypes.c_long(2)
CDF_INT4 = ctypes.c_long(4)
CDF_INT8 = ctypes.c_long(8)
CDF_UINT1 = ctypes.c_long(11)
CDF_UINT2 = ctypes.c_long(12)
CDF_UINT4 = ctypes.c_long(14)
CDF_REAL4 = ctypes.c_long(21)
CDF_REAL8 = ctypes.c_long(22)
CDF_EPOCH = ctypes.c_long(31)
CDF_EPOCH16 = ctypes.c_long(32)
CDF_TIME_TT2000 = ctypes.c_long(33)
CDF_BYTE = ctypes.c_long(41)
CDF_FLOAT = ctypes.c_long(44)
CDF_DOUBLE = ctypes.c_long(45)
CDF_CHAR = ctypes.c_long(51)
CDF_UCHAR = ctypes.c_long(52)
NETWORK_ENCODING = ctypes.c_long(1)
SUN_ENCODING = ctypes.c_long(2)
VAX_ENCODING = ctypes.c_long(3)
DECSTATION_ENCODING = ctypes.c_long(4)
SGi_ENCODING = ctypes.c_long(5)
IBMPC_ENCODING = ctypes.c_long(6)
IBMRS_ENCODING = ctypes.c_long(7)
HOST_ENCODING = ctypes.c_long(8)
PPC_ENCODING = ctypes.c_long(9)
HP_ENCODING = ctypes.c_long(11)
NeXT_ENCODING = ctypes.c_long(12)
ALPHAOSF1_ENCODING = ctypes.c_long(13)
ALPHAVMSd_ENCODING = ctypes.c_long(14)
ALPHAVMSg_ENCODING = ctypes.c_long(15)
ALPHAVMSi_ENCODING = ctypes.c_long(16)
NETWORK_DECODING = NETWORK_ENCODING
SUN_DECODING = SUN_ENCODING
VAX_DECODING = VAX_ENCODING
DECSTATION_DECODING = DECSTATION_ENCODING
SGi_DECODING = SGi_ENCODING
IBMPC_DECODING = IBMPC_ENCODING
IBMRS_DECODING = IBMRS_ENCODING
HOST_DECODING = HOST_ENCODING
PPC_DECODING = PPC_ENCODING
MAC_ENCODING = PPC_ENCODING
MAC_DECODING = PPC_ENCODING
HP_DECODING = HP_ENCODING
NeXT_DECODING = NeXT_ENCODING
ALPHAOSF1_DECODING = ALPHAOSF1_ENCODING
ALPHAVMSd_DECODING = ALPHAVMSd_ENCODING
ALPHAVMSg_DECODING = ALPHAVMSg_ENCODING
ALPHAVMSi_DECODING = ALPHAVMSi_ENCODING
VARY = ctypes.c_long(-1)
NOVARY = ctypes.c_long(0)
ROW_MAJOR = ctypes.c_long(1)
COLUMN_MAJOR = ctypes.c_long(2)
SINGLE_FILE = ctypes.c_long(1)
MULTI_FILE = ctypes.c_long(2)
NO_CHECKSUM = ctypes.c_long(0)
MD5_CHECKSUM = ctypes.c_long(1)
OTHER_CHECKSUM = ctypes.c_long(2)
GLOBAL_SCOPE = ctypes.c_long(1)
VARIABLE_SCOPE = ctypes.c_long(2)
READONLYon = ctypes.c_long(-1)
READONLYoff = ctypes.c_long(0)
VALIDATEFILEon = ctypes.c_long(-1)
VALIDATEFILEoff = ctypes.c_long(0)
zMODEoff = ctypes.c_long(0)
zMODEon1 = ctypes.c_long(1)
zMODEon2 = ctypes.c_long(2)
NEGtoPOSfp0on = ctypes.c_long(-1)
NEGtoPOSfp0off = ctypes.c_long(0)
BACKWARDFILEon = 1
BACKWARDFILEoff = 0
CDF_MAX_PARMS = 5
NO_COMPRESSION = ctypes.c_long(0)
RLE_COMPRESSION = ctypes.c_long(1)
HUFF_COMPRESSION = ctypes.c_long(2)
AHUFF_COMPRESSION = ctypes.c_long(3)
GZIP_COMPRESSION = ctypes.c_long(5)
RLE_OF_ZEROs = ctypes.c_long(0)
OPTIMAL_ENCODING_TREES = ctypes.c_long(0)
NO_SPARSEARRAYS = ctypes.c_long(0)
NO_SPARSERECORDS = ctypes.c_long(0)
PAD_SPARSERECORDS = ctypes.c_long(1)
PREV_SPARSERECORDS = ctypes.c_long(2)
RESERVED_CDFID = 0  # Added by hand (NULL in C)
RESERVED_CDFSTATUS = -1
ILLEGAL_EPOCH_VALUE = ctypes.c_double(-1.0)
ILLEGAL_TT2000_VALUE = ctypes.c_longlong(-9223372036854775805)
FILLED_TT2000_VALUE = ctypes.c_longlong(-9223372036854775807 - 1)
VIRTUAL_RECORD_DATA = 1001
DID_NOT_COMPRESS = 1002
VAR_ALREADY_CLOSED = 1003
SINGLE_FILE_FORMAT = 1004
NO_PADVALUE_SPECIFIED = 1005
NO_VARS_IN_CDF = 1006
MULTI_FILE_FORMAT = 1007
SOME_ALREADY_ALLOCATED = 1008
PRECEEDING_RECORDS_ALLOCATED = 1009
CDF_OK = 0
ATTR_NAME_TRUNC = -1001
CDF_NAME_TRUNC = -1002
VAR_NAME_TRUNC = -1003
NEGATIVE_FP_ZERO = -1004
FORCED_PARAMETER = -1006
NA_FOR_VARIABLE = -1007
CDF_WARN = -2000
ATTR_EXISTS = -2001
BAD_CDF_ID = -2002
BAD_DATA_TYPE = -2003
BAD_DIM_SIZE = -2004
BAD_DIM_INDEX = -2005
BAD_ENCODING = -2006
BAD_MAJORITY = -2007
BAD_NUM_DIMS = -2008
BAD_REC_NUM = -2009
BAD_SCOPE = -2010
BAD_NUM_ELEMS = -2011
CDF_OPEN_ERROR = -2012
CDF_EXISTS = -2013
BAD_FORMAT = -2014
BAD_ALLOCATE_RECS = -2015
BAD_CDF_EXTENSION = -2016
NO_SUCH_ATTR = -2017
NO_SUCH_ENTRY = -2018
NO_SUCH_VAR = -2019
VAR_READ_ERROR = -2020
VAR_WRITE_ERROR = -2021
BAD_ARGUMENT = -2022
IBM_PC_OVERFLOW = -2023
TOO_MANY_VARS = -2024
VAR_EXISTS = -2025
BAD_MALLOC = -2026
NOT_A_CDF = -2027
CORRUPTED_V2_CDF = -2028
VAR_OPEN_ERROR = -2029
BAD_INITIAL_RECS = -2030
BAD_BLOCKING_FACTOR = -2031
END_OF_VAR = -2032
BAD_CDFSTATUS = -2034
CDF_INTERNAL_ERROR = -2035
BAD_NUM_VARS = -2036
BAD_REC_COUNT = -2037
BAD_REC_INTERVAL = -2038
BAD_DIM_COUNT = -2039
BAD_DIM_INTERVAL = -2040
BAD_VAR_NUM = -2041
BAD_ATTR_NUM = -2042
BAD_ENTRY_NUM = -2043
BAD_ATTR_NAME = -2044
BAD_VAR_NAME = -2045
NO_ATTR_SELECTED = -2046
NO_ENTRY_SELECTED = -2047
NO_VAR_SELECTED = -2048
BAD_CDF_NAME = -2049
CANNOT_CHANGE = -2051
NO_STATUS_SELECTED = -2052
NO_CDF_SELECTED = -2053
READ_ONLY_DISTRIBUTION = -2054
CDF_CLOSE_ERROR = -2055
VAR_CLOSE_ERROR = -2056
BAD_FNC_OR_ITEM = -2058
ILLEGAL_ON_V1_CDF = -2060
BAD_CACHE_SIZE = -2063
CDF_CREATE_ERROR = -2066
NO_SUCH_CDF = -2067
VAR_CREATE_ERROR = -2068
READ_ONLY_MODE = -2070
ILLEGAL_IN_zMODE = -2071
BAD_zMODE = -2072
BAD_READONLY_MODE = -2073
CDF_READ_ERROR = -2074
CDF_WRITE_ERROR = -2075
ILLEGAL_FOR_SCOPE = -2076
NO_MORE_ACCESS = -2077
BAD_DECODING = -2079
BAD_NEGtoPOSfp0_MODE = -2081
UNSUPPORTED_OPERATION = -2082
CDF_SAVE_ERROR = -2083
VAR_SAVE_ERROR = -2084
NO_WRITE_ACCESS = -2086
NO_DELETE_ACCESS = -2087
CDF_DELETE_ERROR = -2088
VAR_DELETE_ERROR = -2089
UNKNOWN_COMPRESSION = -2090
CANNOT_COMPRESS = -2091
DECOMPRESSION_ERROR = -2092
COMPRESSION_ERROR = -2093
EMPTY_COMPRESSED_CDF = -2096
BAD_COMPRESSION_PARM = -2097
UNKNOWN_SPARSENESS = -2098
CANNOT_SPARSERECORDS = -2099
CANNOT_SPARSEARRAYS = -2100
TOO_MANY_PARMS = -2101
NO_SUCH_RECORD = -2102
CANNOT_ALLOCATE_RECORDS = -2103
SCRATCH_DELETE_ERROR = -2106
SCRATCH_CREATE_ERROR = -2107
SCRATCH_READ_ERROR = -2108
SCRATCH_WRITE_ERROR = -2109
BAD_SPARSEARRAYS_PARM = -2110
BAD_SCRATCH_DIR = -2111
NOT_A_CDF_OR_NOT_SUPPORTED = -2113
CORRUPTED_V3_CDF = -2223
ILLEGAL_EPOCH_FIELD = -2224
BAD_CHECKSUM = -2225
CHECKSUM_ERROR = -2226
CHECKSUM_NOT_ALLOWED = -2227
IS_A_NETCDF = -2228
TT2000_TIME_ERROR = -2229
UNABLE_TO_PROCESS_CDF = -2230
ZLIB_DECOMPRESSION_ERROR = - 2231
CREATE_ = ctypes.c_long(1001)
OPEN_ = ctypes.c_long(1002)
DELETE_ = ctypes.c_long(1003)
CLOSE_ = ctypes.c_long(1004)
SELECT_ = ctypes.c_long(1005)
CONFIRM_ = ctypes.c_long(1006)
GET_ = ctypes.c_long(1007)
PUT_ = ctypes.c_long(1008)
SAVE_ = ctypes.c_long(1009)
BACKWARD_ = ctypes.c_long(1010)
GETCDFFILEBACKWARD_ = ctypes.c_long(1011)
CHECKSUM_ = ctypes.c_long(1012)
GETCDFCHECKSUM_ = ctypes.c_long(1013)
VALIDATE_ = ctypes.c_long(1014)
GETCDFVALIDATE_ = ctypes.c_long(1015)
GETLEAPSECONDSENVVAR_ = ctypes.c_long(1016)
NULL_ = ctypes.c_long(1000)
CDF_ = ctypes.c_long(1)
CDF_NAME_ = ctypes.c_long(2)
CDF_ENCODING_ = ctypes.c_long(3)
CDF_DECODING_ = ctypes.c_long(4)
CDF_MAJORITY_ = ctypes.c_long(5)
CDF_FORMAT_ = ctypes.c_long(6)
CDF_COPYRIGHT_ = ctypes.c_long(7)
CDF_NUMrVARS_ = ctypes.c_long(8)
CDF_NUMzVARS_ = ctypes.c_long(9)
CDF_NUMATTRS_ = ctypes.c_long(10)
CDF_NUMgATTRS_ = ctypes.c_long(11)
CDF_NUMvATTRS_ = ctypes.c_long(12)
CDF_VERSION_ = ctypes.c_long(13)
CDF_RELEASE_ = ctypes.c_long(14)
CDF_INCREMENT_ = ctypes.c_long(15)
CDF_STATUS_ = ctypes.c_long(16)
CDF_READONLY_MODE_ = ctypes.c_long(17)
CDF_zMODE_ = ctypes.c_long(18)
CDF_NEGtoPOSfp0_MODE_ = ctypes.c_long(19)
LIB_COPYRIGHT_ = ctypes.c_long(20)
LIB_VERSION_ = ctypes.c_long(21)
LIB_RELEASE_ = ctypes.c_long(22)
LIB_INCREMENT_ = ctypes.c_long(23)
LIB_subINCREMENT_ = ctypes.c_long(24)
rVARs_NUMDIMS_ = ctypes.c_long(25)
rVARs_DIMSIZES_ = ctypes.c_long(26)
rVARs_MAXREC_ = ctypes.c_long(27)
rVARs_RECDATA_ = ctypes.c_long(28)
rVARs_RECNUMBER_ = ctypes.c_long(29)
rVARs_RECCOUNT_ = ctypes.c_long(30)
rVARs_RECINTERVAL_ = ctypes.c_long(31)
rVARs_DIMINDICES_ = ctypes.c_long(32)
rVARs_DIMCOUNTS_ = ctypes.c_long(33)
rVARs_DIMINTERVALS_ = ctypes.c_long(34)
rVAR_ = ctypes.c_long(35)
rVAR_NAME_ = ctypes.c_long(36)
rVAR_DATATYPE_ = ctypes.c_long(37)
rVAR_NUMELEMS_ = ctypes.c_long(38)
rVAR_RECVARY_ = ctypes.c_long(39)
rVAR_DIMVARYS_ = ctypes.c_long(40)
rVAR_NUMBER_ = ctypes.c_long(41)
rVAR_DATA_ = ctypes.c_long(42)
rVAR_HYPERDATA_ = ctypes.c_long(43)
rVAR_SEQDATA_ = ctypes.c_long(44)
rVAR_SEQPOS_ = ctypes.c_long(45)
rVAR_MAXREC_ = ctypes.c_long(46)
rVAR_MAXallocREC_ = ctypes.c_long(47)
rVAR_DATASPEC_ = ctypes.c_long(48)
rVAR_PADVALUE_ = ctypes.c_long(49)
rVAR_INITIALRECS_ = ctypes.c_long(50)
rVAR_BLOCKINGFACTOR_ = ctypes.c_long(51)
rVAR_nINDEXRECORDS_ = ctypes.c_long(52)
rVAR_nINDEXENTRIES_ = ctypes.c_long(53)
rVAR_EXISTENCE_ = ctypes.c_long(54)
zVARs_MAXREC_ = ctypes.c_long(55)
zVARs_RECDATA_ = ctypes.c_long(56)
zVAR_ = ctypes.c_long(57)
zVAR_NAME_ = ctypes.c_long(58)
zVAR_DATATYPE_ = ctypes.c_long(59)
zVAR_NUMELEMS_ = ctypes.c_long(60)
zVAR_NUMDIMS_ = ctypes.c_long(61)
zVAR_DIMSIZES_ = ctypes.c_long(62)
zVAR_RECVARY_ = ctypes.c_long(63)
zVAR_DIMVARYS_ = ctypes.c_long(64)
zVAR_NUMBER_ = ctypes.c_long(65)
zVAR_DATA_ = ctypes.c_long(66)
zVAR_HYPERDATA_ = ctypes.c_long(67)
zVAR_SEQDATA_ = ctypes.c_long(68)
zVAR_SEQPOS_ = ctypes.c_long(69)
zVAR_MAXREC_ = ctypes.c_long(70)
zVAR_MAXallocREC_ = ctypes.c_long(71)
zVAR_DATASPEC_ = ctypes.c_long(72)
zVAR_PADVALUE_ = ctypes.c_long(73)
zVAR_INITIALRECS_ = ctypes.c_long(74)
zVAR_BLOCKINGFACTOR_ = ctypes.c_long(75)
zVAR_nINDEXRECORDS_ = ctypes.c_long(76)
zVAR_nINDEXENTRIES_ = ctypes.c_long(77)
zVAR_EXISTENCE_ = ctypes.c_long(78)
zVAR_RECNUMBER_ = ctypes.c_long(79)
zVAR_RECCOUNT_ = ctypes.c_long(80)
zVAR_RECINTERVAL_ = ctypes.c_long(81)
zVAR_DIMINDICES_ = ctypes.c_long(82)
zVAR_DIMCOUNTS_ = ctypes.c_long(83)
zVAR_DIMINTERVALS_ = ctypes.c_long(84)
ATTR_ = ctypes.c_long(85)
ATTR_SCOPE_ = ctypes.c_long(86)
ATTR_NAME_ = ctypes.c_long(87)
ATTR_NUMBER_ = ctypes.c_long(88)
ATTR_MAXgENTRY_ = ctypes.c_long(89)
ATTR_NUMgENTRIES_ = ctypes.c_long(90)
ATTR_MAXrENTRY_ = ctypes.c_long(91)
ATTR_NUMrENTRIES_ = ctypes.c_long(92)
ATTR_MAXzENTRY_ = ctypes.c_long(93)
ATTR_NUMzENTRIES_ = ctypes.c_long(94)
ATTR_EXISTENCE_ = ctypes.c_long(95)
gENTRY_ = ctypes.c_long(96)
gENTRY_EXISTENCE_ = ctypes.c_long(97)
gENTRY_DATATYPE_ = ctypes.c_long(98)
gENTRY_NUMELEMS_ = ctypes.c_long(99)
gENTRY_DATASPEC_ = ctypes.c_long(100)
gENTRY_DATA_ = ctypes.c_long(101)
rENTRY_ = ctypes.c_long(102)
rENTRY_NAME_ = ctypes.c_long(103)
rENTRY_EXISTENCE_ = ctypes.c_long(104)
rENTRY_DATATYPE_ = ctypes.c_long(105)
rENTRY_NUMELEMS_ = ctypes.c_long(106)
rENTRY_DATASPEC_ = ctypes.c_long(107)
rENTRY_DATA_ = ctypes.c_long(108)
zENTRY_ = ctypes.c_long(109)
zENTRY_NAME_ = ctypes.c_long(110)
zENTRY_EXISTENCE_ = ctypes.c_long(111)
zENTRY_DATATYPE_ = ctypes.c_long(112)
zENTRY_NUMELEMS_ = ctypes.c_long(113)
zENTRY_DATASPEC_ = ctypes.c_long(114)
zENTRY_DATA_ = ctypes.c_long(115)
STATUS_TEXT_ = ctypes.c_long(116)
CDF_CACHESIZE_ = ctypes.c_long(117)
rVARs_CACHESIZE_ = ctypes.c_long(118)
zVARs_CACHESIZE_ = ctypes.c_long(119)
rVAR_CACHESIZE_ = ctypes.c_long(120)
zVAR_CACHESIZE_ = ctypes.c_long(121)
zVARs_RECNUMBER_ = ctypes.c_long(122)
rVAR_ALLOCATERECS_ = ctypes.c_long(123)
zVAR_ALLOCATERECS_ = ctypes.c_long(124)
DATATYPE_SIZE_ = ctypes.c_long(125)
CURgENTRY_EXISTENCE_ = ctypes.c_long(126)
CURrENTRY_EXISTENCE_ = ctypes.c_long(127)
CURzENTRY_EXISTENCE_ = ctypes.c_long(128)
CDF_INFO_ = ctypes.c_long(129)
CDF_COMPRESSION_ = ctypes.c_long(130)
zVAR_COMPRESSION_ = ctypes.c_long(131)
zVAR_SPARSERECORDS_ = ctypes.c_long(132)
zVAR_SPARSEARRAYS_ = ctypes.c_long(133)
zVAR_ALLOCATEBLOCK_ = ctypes.c_long(134)
zVAR_NUMRECS_ = ctypes.c_long(135)
zVAR_NUMallocRECS_ = ctypes.c_long(136)
rVAR_COMPRESSION_ = ctypes.c_long(137)
rVAR_SPARSERECORDS_ = ctypes.c_long(138)
rVAR_SPARSEARRAYS_ = ctypes.c_long(139)
rVAR_ALLOCATEBLOCK_ = ctypes.c_long(140)
rVAR_NUMRECS_ = ctypes.c_long(141)
rVAR_NUMallocRECS_ = ctypes.c_long(142)
rVAR_ALLOCATEDFROM_ = ctypes.c_long(143)
rVAR_ALLOCATEDTO_ = ctypes.c_long(144)
zVAR_ALLOCATEDFROM_ = ctypes.c_long(145)
zVAR_ALLOCATEDTO_ = ctypes.c_long(146)
zVAR_nINDEXLEVELS_ = ctypes.c_long(147)
rVAR_nINDEXLEVELS_ = ctypes.c_long(148)
CDF_SCRATCHDIR_ = ctypes.c_long(149)
rVAR_RESERVEPERCENT_ = ctypes.c_long(150)
zVAR_RESERVEPERCENT_ = ctypes.c_long(151)
rVAR_RECORDS_ = ctypes.c_long(152)
zVAR_RECORDS_ = ctypes.c_long(153)
STAGE_CACHESIZE_ = ctypes.c_long(154)
COMPRESS_CACHESIZE_ = ctypes.c_long(155)
CDF_CHECKSUM_ = ctypes.c_long(156)
CDFwithSTATS_ = ctypes.c_long(200)
CDF_ACCESS_ = ctypes.c_long(201)
TT2000END = -99999.999
CDF_DOCUMENT_LEN = CDF_COPYRIGHT_LEN
CDF_ERRTEXT_LEN = CDF_STATUSTEXT_LEN
CDF_NUMDIMS_ = rVARs_NUMDIMS_
CDF_DIMSIZES_ = rVARs_DIMSIZES_
CDF_MAXREC_ = rVARs_MAXREC_
CDF_RECNUMBER_ = rVARs_RECNUMBER_
CDF_RECCOUNT_ = rVARs_RECCOUNT_
CDF_RECINTERVAL_ = rVARs_RECINTERVAL_
CDF_DIMINDICES_ = rVARs_DIMINDICES_
CDF_DIMCOUNTS_ = rVARs_DIMCOUNTS_
CDF_DIMINTERVALS_ = rVARs_DIMINTERVALS_
CDF_NUMVARS_ = CDF_NUMrVARS_
VAR_ = rVAR_
VAR_NAME_ = rVAR_NAME_
VAR_DATATYPE_ = rVAR_DATATYPE_
VAR_NUMELEMS_ = rVAR_NUMELEMS_
VAR_RECVARY_ = rVAR_RECVARY_
VAR_DIMVARYS_ = rVAR_DIMVARYS_
VAR_NUMBER_ = rVAR_NUMBER_
VAR_DATA_ = rVAR_DATA_
VAR_HYPERDATA_ = rVAR_HYPERDATA_
VAR_SEQDATA_ = rVAR_SEQDATA_
VAR_SEQPOS_ = rVAR_SEQPOS_
VAR_MAXREC_ = rVAR_MAXREC_
VAR_DATASPEC_ = rVAR_DATASPEC_
VAR_FILLVALUE_ = rVAR_PADVALUE_
VAR_INITIALRECS_ = rVAR_INITIALRECS_
VAR_EXTENDRECS_ = rVAR_BLOCKINGFACTOR_
ATTR_MAXENTRY_ = ATTR_MAXrENTRY_
ATTR_NUMENTRIES_ = ATTR_NUMrENTRIES_
ENTRY_ = rENTRY_
ENTRY_DATATYPE_ = rENTRY_DATATYPE_
ENTRY_NUMELEMS_ = rENTRY_NUMELEMS_
ENTRY_DATA_ = rENTRY_DATA_
MIPSEL_ENCODING = DECSTATION_ENCODING
MIPSEB_ENCODING = SGi_ENCODING
rVAR_EXISTANCE_ = rVAR_EXISTENCE_
zVAR_EXISTANCE_ = zVAR_EXISTENCE_
ATTR_EXISTANCE_ = ATTR_EXISTENCE_
gENTRY_EXISTANCE_ = gENTRY_EXISTENCE_
rENTRY_EXISTANCE_ = rENTRY_EXISTENCE_
zENTRY_EXISTANCE_ = zENTRY_EXISTENCE_
GLOBAL_SCOPE_ASSUMED = GLOBAL_SCOPE
VARIABLE_SCOPE_ASSUMED = VARIABLE_SCOPE
BAD_EXTEND_RECS = BAD_BLOCKING_FACTOR
rVAR_EXTENDRECS_ = rVAR_BLOCKINGFACTOR_
zVAR_EXTENDRECS_ = zVAR_BLOCKINGFACTOR_
COL_MAJOR = COLUMN_MAJOR
NONE_CHECKSUM = NO_CHECKSUM

# vim: set tw=79 :



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from poppy.core.db.connector import Connector
from poppy.core.db.database import Database
from poppy.core.logger import logger
from poppy.core.tools.imports import import_class
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.exc import NoResultFound, MultipleResultsFound

__all__ = [
    "load_databases",
    "link_databases",
    "yield_limit",
    "connect_databases",
    "get_or_create",
    "get_or_create_with_info",
    "create", "update_one",
    "get_model_dict",
]


def yield_limit(qry, pk_attr, maxrq=1000):
    """
    Specialized windowed query generator (using LIMIT/OFFSET)

    This recipe is to select through a large number of rows thats too
    large to fetch at once. The technique depends on the primary key
    of the FROM clause being an integer value, and selects items
    using LIMIT.
    """
    firstid = None
    while True:
        q = qry
        if firstid is not None:
            q = qry.filter(pk_attr > firstid)
        rec = None
        for rec in q.order_by(pk_attr).limit(maxrq):
            yield rec
        if rec is None:
            break
        firstid = pk_attr.__get__(rec, pk_attr) if rec else None


def load_databases(databases):
    """
    Load databases according to the information provided in the dictionary in
    argument.
    """
    # loop over parameters in the argument
    for database in databases:
        # create the database
        name = database["identifier"]
        logger.debug("Instantiate the database {0}".format(name))
        db = Database(name)

        # set the base used to reflect the database
        if database["identifier"] in Database.bases_manager.availables:
            db.base = Database.bases_manager.availables[database["identifier"]]
        else:
            logger.info("Base of database {0} isn't registered".format(
                database["identifier"]))
            continue

        # update the parameters used by the database
        db.parameters = database["login_info"]


def link_databases(databases):
    """
    Given a list of databases with their properties in a dictionary, create the
    connectors if not already created.
    """
    # loop over databases
    for database in databases:
        # name of the database
        name = database["identifier"]

        # if the connector is already defined
        if name in Connector.manager:
            logger.debug("Connector {0} already instantiated".format(name))
            return

        # get the class for the connector
        if "connector" in database:
            connector = import_class(database["connector"])
        else:
            connector = Connector

        # create an instance of the connector with the name provided
        instance = connector(name)

        # set the database name for the connector
        instance.database = name


def connect_databases():
    """
    Make a connection and reflection of databases registered.
    """
    # loop over databases instances and make the connection
    for database in Database.manager.instances:
        # connect to the database
        database.connectDatabase()

        # if the database is not connected, display it to the user
        if not database.connected:
            logger.error("{0} is not connected".format(database))


def get_or_create(
    session,
    model,
    create_method="",
    create_method_kwargs=None,
    **kwargs
):
    """
    Simply get an object if already present in the database or create it in the
    other case. See
    http://skien.cc/blog/2014/01/15/sqlalchemy-and-race-conditions-implementing/
    and
    http://skien.cc/blog/2014/02/06/sqlalchemy-and-race-conditions-follow-up/
    for better details on why this function as been upgraded to the provided
    example. Better handling of weird cases in the situation of multiple
    processes using the database at the same time.
    """
    try:
        return session.query(model).filter_by(**kwargs).one()
    except NoResultFound:
        kwargs.update(create_method_kwargs or {})
        created = getattr(model, create_method, model)(**kwargs)
        try:
            session.add(created)
            session.commit()
            return created
        except IntegrityError:
            session.rollback()
            return session.query(model).filter_by(**kwargs).one()


def get_update_or_create(
        session,
        model,
        create_method="",
        update_fields=None,
        **kwargs
):
    """
    Get an object if already present in the database and update it. If not, create it
    """
    try:
        # get the obj instance
        instance = session.query(model).filter_by(**kwargs).one()

        # update it
        for field in update_fields or {}:
            setattr(instance, field, update_fields[field])
        session.commit()
        return instance

    except NoResultFound:
        # create the obj instance
        kwargs.update(update_fields or {})
        created = getattr(model, create_method, model)(**kwargs)
        try:
            session.add(created)
            session.commit()
            return created
        except IntegrityError:
            session.rollback()
            return session.query(model).filter_by(**kwargs).one()

def get_or_create_with_info(
    session,
    model,
    create_method="",
    create_method_kwargs=None,
    **kwargs
):
    """
    Simply get an object if already present in the database or create it in the
    other case. See
    http://skien.cc/blog/2014/01/15/sqlalchemy-and-race-conditions-implementing/
    and
    http://skien.cc/blog/2014/02/06/sqlalchemy-and-race-conditions-follow-up/
    for better details on why this function as been upgraded to the provided
    example. Better handling of weird cases in the situation of multiple
    processes using the database at the same time.
    """
    try:
        return session.query(model).filter_by(**kwargs).one(), False
    except NoResultFound:
        kwargs.update(create_method_kwargs or {})
        created = getattr(model, create_method, model)(**kwargs)
        try:
            session.add(created)
            session.commit()
            return created, True
        except IntegrityError:
            session.rollback()
            return session.query(model).filter_by(**kwargs).one(), False


def create(session, model, **kwargs):
    """
    Simply create an object and add it into the database.
    """
    # create it and add it
    instance = model(**kwargs)
    session.add(instance)
    session.commit()
    return instance


def update_one(
    session,
    model,
    values,
    **kwargs
):
    """Update column values of a single entry in the database."""

    # Check if the entry already exists...
    try:
        entry = session.query(model).filter_by(**kwargs).one()
    except NoResultFound:
        # if not return false
        logger.error("NO ENTRY FOUND IN THE DATABASE FOR ({0})!".format(model))
        raise(NoResultFound)
    except MultipleResultsFound:
        # if multiple entries get the first one
        logger.warning(
            "Multiple entries found in the database ({0})!".format(model))
        entry = session.query(model).filter_by(**kwargs).first()

    # Else update column values for the existing entry
    for col in values.keys():
        exec("entry.{0} = values['{0}']".format(col))

    try:
        session.commit()
    except IntegrityError:
        session.rollback()
        return False


def get_model_dict(model, ignore=[]):
    """
    Return dictionary where keywords are the columns of
     the input database table model.

    :param model: database table model
    :param ignore: list of columns to ignore
    :return: dictionary with table column names as keywords (values are empty).
    """
    return dict((column.name, getattr(model, column.name))
                for column in model.__table__.columns if column.name not in ignore)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import collections

__all__ = ["Graph"]


class Graph(object):
    """
    A class to manage the graph that is storing the topology of the graph.
    """
    def __init__(self):
        """
        Init the structure containing nodes and their relation.
        """
        self.graph = collections.defaultdict(set)

    def add_node(self, node, child):
        """
        Add a node to the graph, with a given child.
        """
        self.graph[node].add(child)

    def find_all_paths(self, start, end, path=None):
        """
        Find all existing paths between to nodes.
        """
        if path is None:
            path = []
        path = path + [start]
        if start == end:
            return [path]
        if start not in self.graph:
            return []
        paths = []
        for node in self.graph[start]:
            if node not in path:
                newpaths = self.find_all_paths(node, end, path)
                for newpath in newpaths:
                    paths.append(newpath)
        return paths

    def find_linking_nodes(self, start, end):
        """
        To find all nodes between two nodes. This is the concatenation of all
        paths between the two nodes in argument.
        """
        # get all paths
        paths = self.find_all_paths(start, end)

        # loop over paths to get the set of unique nodes
        nodes = set()
        for path in paths:
            for node in path:
                nodes.add(node)

        return nodes

    def __contains__(self, task):
        return task in self.graph


# vim: set tw=79 :



import requests


def download_file(filepath, url, **kwargs):
    """
    Download a file from the given URL

    :param filepath: path where the file will be downloaded
    :param url: URL of the file to download
    :param stream: (optional) if ``False``, the response content will be immediately downloaded.
    :param auth: (optional) Auth tuple to enable Basic/Digest/Custom HTTP Auth.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return:
    """

    response = requests.get(url, **kwargs)

    # if the response was successful, no Exception will be raised
    response.raise_for_status()

    with open(filepath, 'wb') as out_file:
        for chunk in response:
            out_file.write(chunk)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os.path as osp
from poppy.core.logger import logger
import shutil
import jsonschema
import json
import sys

from poppy.core.generic.dot_dict import DotDict
from poppy.core.generic.metaclasses import ManagerMeta

__all__ = ["Configuration", "ConfigurationNotFound"]




class ConfigurationNotFound(Exception):
    """
    Exception for the case a configuration file is not found.
    """


class Configuration(DotDict, metaclass=ManagerMeta):
    """
    Class to manage parameters from the configuration file easily, retrieve
    them and setting them again.
    """
    def __init__(self, filename=None, schema=None, name=None, *args, **kwargs):
        """
        To set the filename of the configuration file to read.
        """
        # Configure as usual
        super(Configuration, self).__init__(*args, **kwargs)

        # store the file name
        self.filename = filename

        # store the schema
        self.schema = schema

        # set the name used to register the configuration
        self.name = name

    def read(self):
        """
        Read the configuration file in the json format and set the parameters
        inside the instance itself.
        """
        # check the filename is set
        if not self.filename:
            logger.error(
                "The filename must be set into the configuration class"
            )
            return

        # standard verification for the file
        if osp.isfile(self.filename):
            # open the file and read its structure in json
            with open(self.filename, 'rt') as f:
                # put parameters into dictionary
                logger.debug(
                    "Reading parameters from the configuration file " +
                    "{0}".format(self.filename)
                )
                self.update(json.load(f, object_pairs_hook=DotDict))

        else:
            raise ConfigurationNotFound(
                "The configuration file {0} doesn't exist".format(
                    self.filename
                )
            )

        # if the file extends another one, make a merge
        self.extends()

    def read_validate(self):
        """
        Read the configuration file in JSON format and validate with the schema
        provided.
        """
        # read the file as usual
        self.read()

        # check there is a schema
        if not self.schema:
            logger.error(
                "The path to the schema for {0} is not given".format(
                    self.filename
                )
            )
            sys.exit(-1)

        # read the schema
        schema = self.__class__(self.schema)
        schema.read()

        # check the validity of the descriptor against the schema
        jsonschema.validate(
            self,
            schema,
            format_checker=jsonschema.FormatChecker(),
        )

    def write(self):
        """
        Overwrite the content of the configuration file by the parameters
        actually set, and make a copy of the old file before doing it in order
        to preserve some errors when a problem occurred in the application. A
        clean copy of the configuration file should also always be available in
        SVN repository of the program.
        """
        # returns if filename not set
        if not self.filename:
            return

        # start by doing a copy of the old file (preserving mtime and atime)
        logger.debug("Backup the file into {0}".format(self.filename + "~"))
        shutil.copy2(self.filename, self.filename + "~")

        # now dump the parameters inside the json file
        logger.debug("Writing parameters to {0}".format(self.filename))
        with open(self.filename, "w") as f:
            json.dump(self, f, ensure_ascii=False, indent=4)

    def extends(self):
        """
        Check if the file extends another one, read it and make the merge
        between the source file and this one.
        """
        # check presence of the extends keyword
        if "extends" not in self:
            return

        # get the path of the file to extend
        path = self["extends"]

        # check that path is absolute or not
        if not osp.isabs(path):
            # relative path, make it absolute relatively to the current file
            # get path of the current file
            current = osp.abspath(osp.dirname(self.filename))

            # join the path to the current file
            path = osp.abspath(osp.join(current, path))

        # read the parent file
        parent = self.__class__(path)
        parent.read()

        # merge the two dictionaries
        self.merge(parent, self)
        self.update(parent)

    @classmethod
    def merge(cls, destination, source):
        """
        Recursively merge b dictionary into a dictionary.
        """
        # loop over items in the overriding dict
        for key, value in source.items():
            # check the value is a dictionary
            if isinstance(value, dict):
                # get the field with the key in the destination
                node = destination.get(key, {})

                # merge recursively the value
                cls.merge(node, value)
            else:
                # change the value
                destination[key] = value



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from sqlalchemy import Column, ForeignKey
from sqlalchemy.dialects.postgresql import ENUM
import json
import logging


logger = logging.getLogger('sqlalchemy')

__all__ = ['NonNullColumn']


class NonNullColumn(Column):
    """SQLAlchemy column class with NOT NULL constraint by default

    This class inherits sqlalchemy.Column, but has a NOT NULL constraint by
    default.
    It also implements functions needed to generate the
    Database_Description_Document.
    """
    inherit_cache = True

    def __init__(self, *args, nullable=False, descr='', comment='', unique=False, **kwargs):
        """
        Note that the :param comment: is not the same as the parameter 'comment'
        of sqlalchemy.Column.
        """

        self.descr = descr
        self.doc_comment = comment
        self.priority = None

        self.sql_type = list(args).pop(0)

        if self.sql_type.__class__ == ENUM:
            self.sql_type = 'ENUM'


        try:
            if kwargs['primary_key']:
                self.priority = 'PK'
        except KeyError:
            if nullable:
                self.priority = 'N'
            else:
                self.priority = 'NN'

        if self.priority == 'PK':
            self.descr = 'Primary key'

        if unique:
            self.doc_comment = f'Must be unique. {self.doc_comment}'


        # if column is a foreign key, generate the corresponding description
        try:
            fk = list(args).pop(1)
            if fk.__class__ == ForeignKey:
                column = fk._colspec.split('.')
                self.descr = f'FK reference to {column[2]} in the {column[0]}.{column[1]} table'
        except IndexError:
            pass

        super(NonNullColumn, self).__init__(*args,
                                            **kwargs,
                                            nullable=nullable,
                                            unique=unique)

    def __repr__(self):

        infos = {
            'name': f'{self.key}',
            'sql_type': f'{self.sql_type}',
            'description': f'{self.descr}',
            'priority': f'{self.priority}',
            'comment': f'{self.doc_comment}'
        }

        return json.dumps(infos, indent=4)

    def infos(self):

        return {
            'name': f'{self.key}',
            'sql_type': f'{self.sql_type}',
            'description': f'{self.descr}',
            'priority': f'{self.priority}',
            'comment': f'{self.doc_comment}'
        }





#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import inspect
from weakref import WeakSet, WeakKeyDictionary


__all__ = ["Signal"]


class Signal(object):
    def __init__(self):
        self._functions = WeakSet()
        self._methods = WeakKeyDictionary()
        self._activated = True

    def __call__(self, *args, **kargs):

        # call connected functions only if activated
        if self._activated:
            # Call handler functions
            for func in self._functions:
                func(*args, **kargs)

            # Call handler methods
            for obj, funcs in self._methods.items():
                for func in funcs:
                    func(obj, *args, **kargs)

    def connect(self, slot):
        if inspect.ismethod(slot):
            if slot.__self__ not in self._methods:
                self._methods[slot.__self__] = set()

            self._methods[slot.__self__].add(slot.__func__)

        else:
            self._functions.add(slot)

    def disconnect(self, slot):
        if inspect.ismethod(slot):
            if slot.__self__ in self._methods:
                self._methods[slot.__self__].remove(slot.__func__)
        else:
            if slot in self._functions:
                self._functions.remove(slot)

    def clear(self):
        self._functions.clear()
        self._methods.clear()

    def activate(self):
        """
        Activate the signal to emit.
        """
        self._activated = True

    def deactivate(self):
        """
        Deactivate the signal to emit.
        """
        self._activated = False

    def show(self):
        """
        Show all connected slots to the signal.
        """
        return "Methods: {0}\nFunctions: {1}".format(
            dict(self._methods),
            set(self._functions),
        )




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from poppy.core.logger import logger

from poppy.core.generic.signals import Signal

__all__ = ["Manager", "MultipleClassManager"]


class Manager(object):

    """
    A base class to manage objects needing to be registered, and to inform
    other objects of their creation, deletion, etc.
    """

    # keep the list of manager instances to delete their content at the end of the pipeline run
    manager_list = []

    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls, *args, **kwargs)
        Manager.manager_list.append(instance)
        return instance

    def __init__(self):
        """
        Create containers and signals.
        """
        # containers
        self.availables = {}
        self.instances = []
        self.instancesByName = {}

        # signals
        self.added = Signal()
        self.created = Signal()
        self.deleted = Signal()

    def __getitem__(self, name):
        """
        Get an instance from the manager given only its name, as it was a
        dictionary.
        """

        try:
            return self.instancesByName[name]
        except:
            logger.error("POPPY COMPONENT '{0}' IS NOT DEFINED!\n".format(name) +
                         "(Check first in the descriptors and settings.py file)")
            raise ValueError(
                "POPPY COMPONENT '{0}' IS NOT DEFINED!\n".format(name) +
                         "(Check first in the descriptors and settings.py file)"
            )

    def __contains__(self, name):
        """
        Check if the manager contains a given instance by its name.
        """
        return name in self.instancesByName

    def __iter__(self):
        """
        Iter over all the class instances
        """
        return iter(self.instances)

    def add(self, name, cls):
        """
        Add a class to the register.
        """
        # check presence of the class or not
        if name in self.availables:
            logger.error("Class {0} already available".format(name))
            raise ValueError(
                "The class {0} is already defined".format(name)
            )

        # make the class available
        logger.debug("Class {0} now available".format(name))
        self.availables[name] = cls

        # send a signal with the name and the class
        self.added(name, cls)

    def create(self, instance):
        """
        Register created instance.
        """
        # register it
        logger.debug("Registering instance {0}".format(instance))
        self.instances.append(instance)
        if hasattr(instance, "name"):
            self.instancesByName[instance.name] = instance

        # send signal created
        self.created(instance)

    def delete(self, instance):
        """
        Unregister an instance.
        """
        # remove the instance from the manager
        logger.debug(f"Unregistering instance '{instance.__class__.__name__}:{instance.name}'")
        self.instances.remove(instance)
        self.instancesByName.pop(instance.name, None)

        # send a signal that it has been deleted
        self.deleted(instance)

    def delete_all(self):
        """
        Unregister all the instances.
        """
        for instance in list(self.instances):
            self.delete(instance)






class MultipleClassManager(Manager):

    def add(self, name, cls):
        """
        Add a class to the register.
        """
        # make the class available
        logger.debug("Class {0} now available".format(name))
        self.availables[name] = cls

        # send a signal with the name and the class
        self.added(name, cls)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import importlib

__all__ = ["import_class"]


def import_class(path):
    """
    Dynamically import a class from a given path.
    """
    # split the path for the last name, which is the class
    components = path.rsplit(".", 1)

    # import the module
    module = importlib.import_module(components[0])

    # get the class
    return getattr(module, components[-1])




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from poppy.core.logger import logger
from poppy.core.target.base_target import BaseTarget

__all__ = ['PyObjectTarget']

NoDefault = type(
    'NoDefault',
    (object,),
    {
        '__str__': lambda s: 'NoDefault',
        '__repr__': lambda s: 'NoDefault'
    }
)()


class PyObjectTargetException(Exception):
    pass

class EmptyPyObjectTarget(Exception):
    pass


class PyObjectTarget(BaseTarget):
    """
    PyObjectTarget is target class used to pass and track python object through the pipeline
    """

    def __init__(self, *args, value=NoDefault, **kwargs):
        super().__init__(*args, **kwargs)
        self.value = value

    def update(self, *args, value=NoDefault, **kwargs):
        """
        Update the python object target instance with the given args/kwargs

        Updated attributes:
          - value

        :return:
        """

        super().update(*args, **kwargs)

        # update the value only if a value is given
        if value is not NoDefault:
            try:
                if self.value != value:
                    logger.warning(f"Content of PyObjectTarget '{self.id}' was defined multiple times")
            except EmptyPyObjectTarget:
                pass
            finally:
                self.value = value

    @property
    def value(self):
        if self._value is NoDefault:
            raise EmptyPyObjectTarget(f"Content of PyObjectTarget '{self.id}' is not defined")
        else:
            return self._value

    @value.setter
    def value(self, new_value):
        self._value = new_value

    def __iter__(self):
        if not self._is_multi_target:
            raise PyObjectTargetException("Simple targets are not iterable, use the 'many=True' flag to generate "
                                          "multi-target objects")

        for idx, value in enumerate(self.value):
            # TODO: add parametrization of the target id
            yield PyObjectTarget(f"__{self.id}__{idx}__", self._pipeline, value=value)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import sys
import unittest.mock as mock

from poppy.core.conf import settings
from poppy.pop.scripts import import_module_from_path
from sqlalchemy import (create_engine)
from sqlalchemy import event
from sqlalchemy.orm import sessionmaker

# global application scope. create Session class, engine
Session = sessionmaker()


class TaskTestCase:
    """
    The TaskTestCase class is designed to be overridden in derived classes to create unit tests for tasks.

    Example:
        class TestMyPluginTasks(TaskTestCase):
            def test_task1(self):
                # --- initialize the task ---
                from my_plugin import task1

                self.task = task1()

                # --- create some fake data ---
                # (you can use directly pipeline attributes)

                self.task.pipeline.properties.input_dir = 'my_input_dir'
                self.task.pipeline.properties.output  = 'my_output_dir'

                # (...)

                # --- run the task ---

                self.run_task()

                # --- make assertions ---

                # test the result
                assert self.task.pipeline.properties.result == 'my_result'

                # (...)

            def test_task2(self):
                # --- initialize the task ---
                from my_plugin import task2

                self.task = task2()

                # (...)
    """

    def setUp(self):
        """
        Setup the pipeline before each test
        :return:
        """

        # reset the task
        self._task = None

    def run_task(self):
        if self.task is None:
            raise ValueError('The task has not been initialized')
        else:
            self.task.run()

    @property
    def task(self):
        return self._task

    @task.setter
    def task(self, task):
        # store the task
        self._task = task

        # mock the task to provide access to the pipeline property
        self._task.pipeline = mock.MagicMock()


class CommandTestCase:
    """
    The CommandTestCase class is designed to be overridden in derived classes to create unit tests for commands.

    Example:
        class TestMyPluginCommands(CommandTestCase):
            def test_sub_command1(self):

                # --- create some fake data ---

                value1 = 'value1'
                value2 = 'value2'

                # --- initialize the command ---

                command = ['pop',
                           'my_command',
                           'my_sub_command',
                           '--my_option1', value1,
                           '--my_option2', value2,
                           '--dry-run']

                # --- run the command ---

                self.run_command(command)

                # --- make assertions ---

                # test the result
                assert sorted(['file1', 'file2']) == sorted(os.listdir('my_result_path'))

                # (...)
    """

    # The absolute path to the configuration file used in the database setup
    config_path = None

    # The path to the settings file used in the database setup
    settings_path = None

    # The test database identifier
    database_identifier = None

    def mock_configure_settings(self, module=None, dictionary={}, ignore_pipeline_settings=False):

        # create a mock for the configure method
        def configure(instance, pipeline_settings):

            # ignore the pipeline settings if needed
            if not ignore_pipeline_settings:
                if isinstance(pipeline_settings, dict):
                    instance.from_dict(pipeline_settings)
                else:
                    instance.from_module(pipeline_settings)

            # load custom settings from module
            if module:
                instance.from_module(module)

            # load from dictionary
            if dictionary:
                instance.from_dict(dictionary)

        return configure

    @property
    def settings(self):
        from poppy.core.conf import settings as _settings
        return _settings

    def setup_method(self, method):
        """
        Setup the database before each test
        :return:
        """
        self.setup_database()

        # TODO: handle custom settings
        # using something like @mock.patch('poppy.core.conf.settings.PLUGINS', ['my.plugin1', 'my.plugin2'])
        # TODO: handle custom config

    def teardown_method(self, method):
        self.session.close()

        # rollback - everything that happened with the
        # Session above (including calls to commit())
        # is rolled back.
        self.transaction.rollback()

        # return connection to the Engine
        self.connection.close()

    def generate_url(self, **kwargs):
        """
        Generate the URL of the database from the parameters.
        """
        return '{vendor}://{admin}@{address}/{database}'.format(
            **kwargs
        )

    def load_settings(self):

        if self.settings_path is not None:
            settings_module = import_module_from_path('settings_module', self.settings_path)
        elif os.environ.get('PIPELINE_SETTINGS_FILE', None):
            settings_module = import_module_from_path('settings_module', os.environ['PIPELINE_SETTINGS_FILE'])
        else:
            try:
                import settings as settings_module
            except ModuleNotFoundError:
                # search for settings in the pipeline root dir
                settings_module_path = os.path.join(settings.ROOT_DIRECTORY, 'settings.py')

                if os.path.isfile(settings_module_path):
                    settings_module = import_module_from_path('settings_module', settings_module_path)
                else:
                    raise Exception('No settings file found')

        # update the default settings
        settings.configure(settings_module)

    def load_configuration(self):
        from poppy.core.configuration import Configuration

        if self.config_path is not None:
            self.configuration = Configuration(self.config_path)
        else:
            self.configuration = Configuration(os.getenv('PIPELINE_CONFIG_FILE', None))

        self.configuration.read()

    def setup_database(self):
        """
        Setup the database before the generation of the tested command

        :return:
        """

        # load settings and configuration
        self.load_configuration()
        self.load_settings()

        self.database_identifier = settings.TEST_DATABASE
        database_identifier = getattr(self, 'database_identifier', settings.TEST_DATABASE)

        database_info = list(filter(lambda database: database['identifier'] == database_identifier,
                                    self.configuration['pipeline.databases']))[0]

        url = self.generate_url(**database_info['login_info'])

        self.engine = create_engine(url)

        # connect to the database
        self.connection = self.engine.connect()

        # begin a non-ORM transaction
        self.transaction = self.connection.begin()

        # bind an individual Session to the connection
        self.session = Session(bind=self.connection)

        # start the session in a SAVEPOINT...
        self.session.begin_nested()

        # then each time that SAVEPOINT ends, reopen it
        @event.listens_for(self.session, 'after_transaction_end')
        def restart_savepoint(session, transaction):
            if transaction.nested and not transaction._parent.nested:
                # ensure that state is expired the way
                # session.commit() at the top level normally does
                # (optional step)
                session.expire_all()

                session.begin_nested()

    def mock_create_engine(self, database, *args, **kwargs):
        database.engine = self.engine

    def mock_create_connection(self, database, *args, **kwargs):
        database.db_connection = self.connection

    def mock_session_factory(self):
        return self.session

    def mock_configure(self):
        pass

    def run_command(self, command):
        from poppy.pop.scripts import main as pipeline_main
        # check if the command argument is a list and if it contains the 'pop' keyword
        if not isinstance(command, (list,)):
            try:
                arg_list = command.split()
            except:
                raise TypeError('The command attribute must be a list or a string')
        else:
            arg_list = command

        if next(iter(arg_list), None) != 'pop':
            raise ValueError('The command must start with "pop"')

        # apply mocks on the pipeline main function
        @mock.patch.object(sys, 'argv', arg_list)
        @mock.patch('poppy.core.db.database.Database.scoped_session', new_callable=mock.PropertyMock)
        @mock.patch('poppy.core.db.database.Database.configure', side_effect=self.mock_configure)
        @mock.patch('poppy.core.db.database.Database.create_engine', side_effect=self.mock_create_engine, autospec=True)
        @mock.patch('poppy.core.db.database.Database.create_connection', side_effect=self.mock_create_connection,
                    autospec=True)
        def main(mock_create_connection, mock_create_engine, mock_configure, mock_scoped_session):
            mock_scoped_session.return_value = self.session
            pipeline_main()

        # call the mocked main
        main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import datetime
import sys
import uuid

from poppy.core.db.dry_runner import DryRunner
from poppy.core.db.handlers import get_or_create_with_info
from poppy.core.generic.signals import Signal
from poppy.core.logger import logger
from poppy.pop.models.job import JobLog


class BaseTask:
    """
    Base class for tasks to run through the poppy pipeline.
    """
    # default value for the time of jobs when not started
    default_time = datetime.datetime(1900, 1, 1)

    # status for the task
    PENDING = "Pending"
    STOPPED = "Terminated"
    STARTED = "Running"
    STATE = {PENDING, STOPPED, STARTED}

    OK = "OK"
    WARNING = "WARNING"
    ERROR = "ERROR"
    STATUS = {OK, WARNING, ERROR}

    def __init__(
            self, *,
            plugin=None,
            descriptor=None,
            instance_name=None
    ):
        """
        Initialization of some parameters.
        """
        # store the instance name if any
        self.instance_name = instance_name

        # store the pipeline instance
        self.pipeline = None

        # init the list of parents and children tasks
        self.parents = []
        self.children = []

        # a signal to indicate that the job has changed
        self.changed = Signal()

        # an other one to indicate that it has been created
        self.created = Signal()

        # a signal to indicate that the task as started and ended
        self.started = Signal()
        self.ended = Signal()

        # signal when the task is reset
        self.reseted = Signal()

        # a signal to indicate that an error occurred
        self.errored = Signal()

        # set some reusable properties to their default values
        self.reset()

        # store the the plugin of the job
        self.plugin = plugin

        # store the category of the task
        self.category = descriptor['category']

        # store the textual description
        self.description = descriptor['description']

        # init the targets that can be created from task to None
        self.targets = None

    @property
    def instance_name(self):
        """
        Return the instance name if defined otherwise return the task name

        :return: the instance name or the task name
        """
        if self._instance_name:
            return self._instance_name
        else:
            return self.name

    @instance_name.setter
    def instance_name(self, value):
        self._instance_name = value

    def reset(self):
        """
        To reset some attributes of the task, in order to have one new task
        from an existing one.
        """
        logger.debug("Reset task {0}".format(self))
        # create an uuid for this job
        self.uuid = uuid.uuid4()

        # for the state of the task
        self.completed = False
        self.failed = False

        # set the default start and end times
        self.starttime = self.default_time
        self.endtime = self.default_time

        # default values for job state/status
        self._state = self.PENDING
        self._status = self.OK

        # send a signal that the task has been reset
        self.reseted(self)

    def run(self):
        """
        The method that the other tasks must override to perform their
        computations.
        """
        pass

    def input(self):
        """
        Return a list of the input targets necessary to this task. The list
        contains the names of the attributes used for storing targets into the
        properties object of the pipeline. The task should have a reference on
        it.
        """
        return []

    def output(self):
        """
        Should return a structure containing the targets outputs of these task.
        If a task use them as input, it should know the structure to access to
        good properties.
        """
        return []

    def start(self):
        """
        Indicate at which time the job started.
        """
        # indicate that the task started
        logger.debug("Starting task {0}".format(self))

        # store the time
        self.starttime = datetime.datetime.now()

        # indicate that it is in progress
        self.progress()

        # indicate the change
        self.job_changed()

        # indicate the job as started
        self.started(self)

    def stop(self):
        """
        Indicate at which time the job ended.
        """
        # indicate that the task is stopping
        logger.debug("Stopping task {0}".format(self))

        # store the time
        self.endtime = datetime.datetime.now()

        # indicate terminated
        self.terminated()

        # indicate the change
        self.job_changed()

        # send signal ended
        self.ended(self)

    def job(self):
        """
        Create the job representation of the task into the POPPy database.
        """

        # create the representation in the database
        logger.debug("Get/create representation for job {0}".format(self))
        job, created = get_or_create_with_info(
            self.pipeline.db.session,
            JobLog,
            job_name=repr(self),
            job_uuid=str(self.uuid),
            job_category=self.category,
            job_descr=self.description,
            job_task=self.instance_name,
            job_plugin=self.plugin.name,
            create_method_kwargs=dict(
                job_starttime=self.starttime,
                job_endtime=self.endtime,
                job_state=self._state,
                job_status=self._status,
                job_status_descr="",
            ),
        )

        # indicate that the representation has been created
        if created:
            logger.debug("Created representation for job {0}".format(self))
            self.created(job)

        # store the job
        return job

    def __or__(self, child):
        """
        To set the task in the other side of the pipe as a child task.
        """
        # add the task itself as a parent of the child class
        child.parents.append(self)

        # add the child
        self.children.append(child)

        # return the child, since it will be used to create branching
        return child

    def exception(self, message,
                  exception_type='ERROR',
                  exception_level='Critical'):
        """
        Create an exception in the pipeline, i.e. in the POPPy database with the
        given message as description.
        """
        self.pipeline.create_exception(self, message,
                                       exception_type=exception_type,
                                       exception_level=exception_level)

    @DryRunner.dry_run
    def job_changed(self):
        """
        What to do when the status of the job has changed.
        """
        # get the job in the database
        job = self.job()

        # store the new information on the representation
        job.job_starttime = self.starttime
        job.job_endtime = self.endtime
        job.job_state = self._state
        job.job_status = self._status
        job.parent = self.parents[0].job() if len(self.parents) else None

        # commit changed
        self.pipeline.db.update_database()

        # indicate that the task job has changed
        self.changed()

    def state(self, value):
        """
        To change the state of the task job with the given provided argument.
        """
        # check that the value is accepted
        if value not in self.STATE:
            raise ValueError(
                "The state of the task job must be set to {0}".format(self.STATE)
            )

        # store the state
        logger.debug("Task job {0} is now at state {1}".format(self, value))
        self._state = value

        # call internal method to change the state of a target
        self.job_changed()

    def status(self, value):
        """
        To change the status of the target.
        """
        # check that the value is accepted
        if value not in self.STATUS:
            raise ValueError(
                "The status of the task job must be set to " +
                "{0}".format(self.STATUS)
            )

        # store the status
        logger.debug(
            "Task job {0} is now at status {1}".format(self, value)
        )
        self._status = value

        # indicate the status changed
        self.job_changed()

    def ok(self):
        """
        Set to ok the status
        """
        self.status(self.OK)

    def warning(self):
        """
        Set the status to warning.
        """
        self.status(self.WARNING)

    def error(self):
        """
        Set the status to error.
        """
        # store the time
        self.endtime = datetime.datetime.now()

        # flag to indicate an error happened
        self.failed = True

        # indicate terminated
        self.terminated()

        # change the status
        self.status(self.ERROR)

        # send signal error happened
        self.errored(self)

    def pending(self):
        """
        Set the state to pending.
        """
        self.state(self.PENDING)

    def terminated(self):
        """
        Set the state to terminated.
        """
        self.state(self.STOPPED)

    def progress(self):
        """
        Set the state to in progress.
        """
        self.state(self.STARTED)

    def descendants(self):
        """
        Give the children-descendants of a task. Those tasks are dependant of
        this task to be run. By default, it returns the list of children as an
        iterable, but it can be transformed into a generator, to be able to
        create loops in the pipeline, returning at a given pint of the pipeline
        according to some specific conditions.
        """
        # returns the child tasks by default
        for child in self.children:
            yield child

    def ancestors(self):
        """
        Equivalent to the descendants methods, but for the parents of the task.
        Maybe an utility will be find some day with it...
        """
        # returns the child tasks by default
        for parent in self.parents:
            yield parent

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, description):
        if description is not None:
            self._description = description
        elif description is None and not hasattr(self, "_description"):
            logger.error(
                "There must be a description for task {0}".format(self)
            )
            sys.exit(-1)

    @property
    def category(self):
        return self._category

    @category.setter
    def category(self, category):
        if category is not None or not hasattr(self, "_category"):
            self._category = category
        elif category is None and not hasattr(self, "_category"):
            logger.error(
                "There must be a category for task {0}".format(self)
            )
            sys.exit(-1)

    @property
    def children(self):
        return self._children

    @children.setter
    def children(self, children):
        self._children = children

    @property
    def parents(self):
        return self._parents

    @parents.setter
    def parents(self, parents):
        self._parents = parents

    def __repr__(self):
        """
        Represent the task with a name. Use both the task and instance name, except if the instance name is
        not defined.
        """
        # use the name if provided
        if self._instance_name:
            return f'{self.name}: {self._instance_name}'
        else:
            return self.name



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os.path as osp
import collections
import importlib
from poppy.core.logger import logger

from poppy.core.generic.metaclasses import ManagerMeta
from poppy.core.configuration import Configuration
from poppy.core.tools.exceptions import print_exception
from poppy.core.generic.cache import CachedProperty
from poppy.core.generic.paths import Paths
from poppy.pop.tools import paths
from poppy.core.task import Task


__all__ = ["Plugin"]


class PluginError(Exception):
    """
    Errors related to plugin loading and definitions.
    """


class Plugin(object, metaclass=ManagerMeta):
    """
    Class to manage plugins module of the pipeline.
    """
    def __init__(self, module_name, module):
        """
        Store the information relative to plugins. Plugins should expose a
        tools module containing a from_config method which returns the absolute
        path to the configuration directory of the plugin with the name of the
        files or directory joined in arguments.
        """
        # store the name of the module
        self.module_name = module_name

        # store the module
        self.module = module


        # load the descriptor and store the name
        self.name = self.descriptor["identification.identifier"]

    def _check_module(self, module):
        """
        Check if a module inside the plugin is importable.
        """
        # name for loading
        name = ".".join([self.module_name, module])

        # first try to import a base for models that will be registered
        if importlib.util.find_spec(name):
            try:
                importlib.import_module(name)
            except ImportError as e:
                print_exception(
                    "Can't load {0} for plugin {1}:".format(
                        module,
                        self.name,
                    )
                )
                raise e

    def get_task_descriptor(self, name):

        tasks = self.tasks

        if name not in tasks:
            raise PluginError(
                "Task {0} not defined in plugin {1} descriptor".format(
                    name,
                    self.name,
                )
            )

        # and to the task information
        return tasks[name]


    def task(self, task_name):
        """
        Create a task from the definition and properties defined in the
        descriptor of the plugin.
        """
        # check that the task is defined in the descriptor
        tasks = self.tasks
        if task_name not in tasks:
            raise PluginError(
                "Task {0} not defined in plugin {1} descriptor".format(
                    task_name,
                    self.name,
                )
            )

        # create the task with the right properties
        return type(
            Task.to_camel_case(task_name),
            (Task,),
            {'plugin_name': self.name, 'name': task_name}
        )

    @CachedProperty
    def tasks(self):
        """
        Create a dictionary of tasks from their description in the plugin
        descriptor.
        """
        # container for tasks
        tasks_data = collections.defaultdict(dict)

        # get the tasks from the descriptor
        descriptor = self.descriptor
        tasks = descriptor["tasks"]

        # loop over tasks in the descriptor
        for task in tasks:
            # add the task
            storage = tasks_data[task["name"]]
            storage["name"] = task["name"]
            storage["category"] = task["category"]
            storage["description"] = task["description"]
            storage["inputs"] = list(task["inputs"].keys())
            storage["outputs"] = list(task["outputs"].keys())
            data = {}
            for name, target in task["inputs"].items():
                data[name] = {
                    "identifier": target["identifier"]
                }
            storage["input_targets"] = data
            data = {}
            for name, target in task["outputs"].items():
                data[name] = {
                    "identifier": target["identifier"]
                }
            storage["output_targets"] = data

        return tasks_data

    @CachedProperty
    def descriptor(self):
        """
        Return the dictionary of the contents of the descriptor of the plugin.
        """
        # check that the path for the descriptor exists
        # get descriptor_path
        descriptor_path = self.paths.from_root(
            "descriptor.json"
        )
        if not osp.isfile(descriptor_path):
            raise PluginError(
                "Cannot access to the descriptor of plugin {0}: {1}".format(
                    self.module_name, descriptor_path,
                )
            )
        else:
            self.descriptor_path = descriptor_path

        # try to read the JSON file
        try:
            data = Configuration(
                descriptor_path,
                paths.from_json_schemas("plugin-descriptor-schema.json"),
            )
            data.read_validate()
        except Exception as e:
            raise PluginError(
                "{0}\nCannot read JSON descriptor of plugin {1}".format(
                    str(e),
                    self.module_name,
                )
            )

        # all is good return data that will be stored in the descriptor
        return data

    @CachedProperty
    def identifier(self):
        return self.descriptor["identification.identifier"]

    @CachedProperty
    def version(self):
        return self.descriptor["release.version"]

    @property
    def module(self):
        return self._module

    @module.setter
    def module(self, module):
        """
        Store the module and check for the good interface of the plugin.
        """
        # keep trace of the module
        self._module = module

        # check that the module is a package
        if not hasattr(self.module, "__path__"):
            raise PluginError(
                "Plugin {0} is not a package".format(self.name)
            )

        # reference the module paths in the plugin (the __path__ contains
        # several paths for importing subpackages, but the first if the package
        # is not doing fancy things should be the root directory of the
        # package, so we only keep the first one as root)
        self.paths = Paths(self.module.__path__[0])

        # load the tasks definitions from the descriptor
        self.tasks

    def load(self):
        """
        Check the integrity of the module after all the rest has been done.
        Imports also module for registration of some commands, tasks, etc.
        """
        # make some verification about the plugin
        # import models
        self._check_module("models")

        # import commands
        self._check_module("commands")

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    def __repr__(self):
        return "Plugin(name={0})".format(self.name)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging


__all__ = ["TerminalColorFormatter"]


BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = tuple(range(8))

# The background is set with 40 plus the number of the color,
# and the foreground with 30

# These are the sequences need to get colored ouput
RESET_SEQ = "\033[0m"
COLOR_SEQ = "\033[%dm"
BOLD_SEQ = "\033[1m"


def formatter_message(message, use_color=True):
    if use_color:
        message = message.replace("$RESET", RESET_SEQ).replace(
            "$BOLD",
            BOLD_SEQ
        )
    else:
        message = message.replace("$RESET", "").replace("$BOLD", "")
    return message

COLORS = {
    'WARNING': YELLOW,
    'INFO': GREEN,
    'DEBUG': BLUE,
    'CRITICAL': WHITE,
    'ERROR': RED
}


class TerminalColorFormatter(logging.Formatter):
    def format(self, record):
        levelname = record.levelname
        text = super(TerminalColorFormatter, self).format(record)
        if record.levelname in COLORS:
            return COLOR_SEQ % (
                30 + COLORS[levelname]
            ) + text + RESET_SEQ
        else:
            return text



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import traceback
import logging

from poppy.core.logger import logger
import sys


__all__ = ["DescriptorLoadError",
            "MissingArgument",
           "MissingProperty",
           "MissingInput",
           "TargetFileNotSaved",
            "print_exception",
            "exception_hook"]

def exception_hook(exctype, value, traceback):
    """Tailored version of the excepthook
    can be used to only print full traceback in DEBUG mode.
    To activate it, add the line sys.excepthook = exception_hook in the code.
    See https://stackoverflow.com/questions/6598053/python-global-exception-handling/6598286#6598286"""
    if logger.getEffectiveLevel() == logging.DEBUG:
        sys.__excepthook__(exctype, value, traceback)

class DescriptorLoadError(Exception):
    """Exception raised when a descriptor file cannot be loaded."""
    def __init__(self, message, *args, **kwargs):
        super(DescriptorLoadError, self).__init__(*args, **kwargs)
        logger.error(message)
        self.message = message

    pass

class MissingInput(Exception):
    """Exception raised if an input target is missing."""
    def __init__(self, message, *args, **kwargs):
        super(MissingInput, self).__init__(*args, **kwargs)
        logger.error(message)
        self.message = message

    pass


class MissingArgument(Exception):
    """
    Exception raised when input argument is missing
    """
    def __init__(self, message, *args, **kwargs):
        super(MissingArgument, self).__init__(*args, **kwargs)
        logger.error(message)
        self.message = message

    pass

class MissingProperty(Exception):
    """Exception raised if a pipeline property is missing."""

    def __init__(self, message, *args, **kwargs):
        super(MissingProperty, self).__init__(*args, **kwargs)
        logger.error(message)
        self.message = message

    pass


class TargetFileNotSaved(Exception):
    """Exception raised if target file not saved correctly."""
    def __init__(self, message, *args, **kwargs):
        super(TargetFileNotSaved, self).__init__(*args, **kwargs)
        logger.error(message)
        self.message = message

    pass

def print_exception(message=None, log_level='error', use_traceback=True):
    """
    To handle the printing of the exception message when one occurred. Particularly
    useful when catching errors and want to add debug information on the same
    time.
    """
    # set the available levels

    logging_function_dict = {'info': logger.info,
                             'warning': logger.warning,
                             'error': logger.error,
                             'critical': logger.critical,
                             'exception': logger.exception, }

    logging_function = logging_function_dict.get(log_level, logger.error)

    if use_traceback:
        # get the traceback
        trace = traceback.format_exc()
    else:
        trace=''

    # if not message provided, get the traceback of errors to be a little more
    # useful for the developer
    if message is not None:
        mess = "\n".join([trace, message])
    else:
        # else use message provided by developer
        mess = trace

    # show error in the logger
    logging_function(mess)

    # return the message
    return mess